PFX_PLATE_DISCIPLINE_METRICS = {
    "Z%": ("zone_rate", "total_inside_strike_zone"),
    "SS%": ("swinging_strike_rate", "total_swinging_strikes"),
    "O-SW%": ("o_swing_rate", "total_swings_outside_zone"),
    "CON%": ("contact_rate", "total_swings_made_contact"),
}

PFX_BATTED_BALL_METRICS = {
    "GB%": ("ground_ball_rate", "total_ground_balls"),
    "LD%": ("line_drive_rate", "total_line_drives"),
    "FB%": ("fly_ball_rate", "total_fly_balls"),
    "PU%": ("pop_up_rate", "total_pop_ups"),
}
