"""Format strings used with strftime/strptime methods."""

DATE_ONLY = "%Y-%m-%d"
DATE_ONLY_2 = "%m/%d/%Y"
DATE_ONLY_TABLE_ID = "%Y%m%d"
DT_NAIVE_LESS = "%m/%d/%Y %I:%M %p"
DT_NAIVE = "%m/%d/%Y %I:%M:%S %p"
DT_AWARE = "%m/%d/%Y %I:%M:%S %p %z"
DT_STR_FORMAT = "%m/%d/%Y %I:%M %p %Z%z"
DT_STR_FORMAT_ALL = "%Y-%m-%d %H:%M:%S.%f %Z%z"
DATE_MONTH_NAME = "%b %d %Y"
FILE_TIMESTAMP = "%Y%m%d_%H%M%S"
CSV_UTC = "%Y-%m-%d %H:%M:%S"
DT_AWARE_VERBOSE = "%A, %B %d %Y %I:%M %p (%Z)"
ISO_8601 = "%Y-%m-%dT%H:%M:%S%z"
HTTP_TIME = "%a, %d %b %Y %H:%M:%S GMT"
