# flake8: noqa
from vigorish.cli.menu_items.admin_tasks.add_to_database import AddToDatabase
from vigorish.cli.menu_items.admin_tasks.backup_database import BackupDatabase
from vigorish.cli.menu_items.admin_tasks.calculate_avg_pitch_times import CalculatePitchTimes
from vigorish.cli.menu_items.admin_tasks.import_scraped_data import ImportScrapedData
from vigorish.cli.menu_items.admin_tasks.npm_install_update import NpmInstallUpdate
from vigorish.cli.menu_items.admin_tasks.restore_database import RestoreDatabase
from vigorish.cli.menu_items.admin_tasks.sync_scraped_data import SyncScrapedData
from vigorish.cli.menu_items.admin_tasks.update_player_id_map import UpdatePlayerIdMap
